@@ -3,6 +3,7 @@ import json
 import time
 from typing import Dict, List, Union
 
+import requests
 import websocket
 
 from pycardano.address import Address
@@ -25,10 +26,17 @@ __all__ = ["OgmiosChainContext"]
 
 
 class OgmiosChainContext(ChainContext):
-    def __init__(self, ws_url: str, network: Network, compact_result=True):
+    def __init__(
+        self,
+        ws_url: str,
+        network: Network,
+        compact_result=True,
+        kupo_url=None,
+    ):
         self._ws_url = ws_url
         self._network = network
         self._service_name = "ogmios.v1:compact" if compact_result else "ogmios"
+        self._kupo_url = kupo_url
         self._last_known_block_slot = 0
         self._genesis_param = None
         self._protocol_param = None
@@ -152,8 +160,34 @@ class OgmiosChainContext(ChainContext):
         args = {"query": "chainTip"}
         return self._request(method, args)["slot"]
 
-    def utxos(self, address: str) -> List[UTxO]:
-        """Get all UTxOs associated with an address.
+    def _extract_asset_info(self, asset_hash: str):
+        policy_hex, asset_name_hex = asset_hash.split(".")
+        policy = ScriptHash.from_primitive(policy_hex)
+        asset_name_hex = AssetName.from_primitive(asset_name_hex)
+
+        return policy_hex, policy, asset_name_hex
+
+    def _check_utxo_unspent(self, tx_id: str, index: int) -> bool:
+        """Check whether an UTxO is unspent with Ogmios.
+
+        Args:
+            tx_id (str): transaction id.
+            index (int): transaction index.
+        """
+
+        method = "Query"
+        args = {"query": {"utxo": [{"txId": tx_id, "index": index}]}}
+        results = self._request(method, args)
+
+        if results:
+            return True
+        else:
+            return False
+
+    def _utxos_kupo(self, address: str) -> List[UTxO]:
+        """Get all UTxOs associated with an address with Kupo.
+        Since UTxO querying will be deprecated from Ogmios in next
+        major release: https://ogmios.dev/mini-protocols/local-state-query/.
 
         Args:
             address (str): An address encoded with bech32.
@@ -161,6 +195,68 @@ class OgmiosChainContext(ChainContext):
         Returns:
             List[UTxO]: A list of UTxOs.
         """
+        address_url = self._kupo_url + "/" + address
+        results = requests.get(address_url).json()
+
+        utxos = []
+
+        for result in results:
+            tx_id = result["transaction_id"]
+            index = result["output_index"]
+
+            # Right now, all UTxOs of the address will be returned with Kupo, which requires Ogmios to
+            # validate if the UTxOs are spent with output reference. This feature is being considered to
+            # be added to Kupo to avoid extra API calls.
+            # See discussion here: https://github.com/CardanoSolutions/kupo/discussions/19.
+            if self._check_utxo_unspent(tx_id, index):
+                tx_in = TransactionInput.from_primitive([tx_id, index])
+
+                lovelace_amount = result["value"]["coins"]
+
+                datum_hash = (
+                    DatumHash.from_primitive(result["datum_hash"])
+                    if result["datum_hash"]
+                    else None
+                )
+
+                if not result["value"]["assets"]:
+                    tx_out = TransactionOutput(
+                        Address.from_primitive(address),
+                        amount=lovelace_amount,
+                        datum_hash=datum_hash,
+                    )
+                else:
+                    multi_assets = MultiAsset()
+
+                    for asset, quantity in result["value"]["assets"].items():
+                        policy_hex, policy, asset_name_hex = self._extract_asset_info(
+                            asset
+                        )
+                        multi_assets.setdefault(policy, Asset())[
+                            asset_name_hex
+                        ] = quantity
+
+                    tx_out = TransactionOutput(
+                        Address.from_primitive(address),
+                        amount=Value(lovelace_amount, multi_assets),
+                        datum_hash=datum_hash,
+                    )
+                utxos.append(UTxO(tx_in, tx_out))
+            else:
+                continue
+
+        return utxos
+
+    def _utxos_ogmios(self, address: str) -> List[UTxO]:
+        """Get all UTxOs associated with an address with Ogmios.
+
+        Args:
+            address (str): An address encoded with bech32.
+
+        Returns:
+            List[UTxO]: A list of UTxOs.
+        """
+
         method = "Query"
         args = {"query": {"utxo": [address]}}
         results = self._request(method, args)
@@ -187,11 +283,8 @@ class OgmiosChainContext(ChainContext):
             else:
                 multi_assets = MultiAsset()
 
-                for asset in output["value"]["assets"]:
-                    quantity = output["value"]["assets"][asset]
-                    policy_hex, asset_name_hex = asset.split(".")
-                    policy = ScriptHash.from_primitive(policy_hex)
-                    asset_name_hex = AssetName.from_primitive(asset_name_hex)
+                for asset, quantity in output["value"]["assets"].items():
+                    policy_hex, policy, asset_name_hex = self._extract_asset_info(asset)
                     multi_assets.setdefault(policy, Asset())[asset_name_hex] = quantity
 
                 tx_out = TransactionOutput(
@@ -203,6 +296,22 @@ class OgmiosChainContext(ChainContext):
 
         return utxos
 
+    def utxos(self, address: str) -> List[UTxO]:
+        """Get all UTxOs associated with an address.
+
+        Args:
+            address (str): An address encoded with bech32.
+
+        Returns:
+            List[UTxO]: A list of UTxOs.
+        """
+        if self._kupo_url:
+            utxos = self._utxos_kupo(address)
+        else:
+            utxos = self._utxos_ogmios(address)
+
+        return utxos
+
     def submit_tx(self, cbor: Union[bytes, str]):
         """Submit a transaction to the blockchain.
 
