@@ -2,10 +2,17 @@ from __future__ import annotations
 
 from copy import copy, deepcopy
 from dataclasses import dataclass, field, fields
-from typing import Dict, List, Optional, Set, Union
+from typing import Dict, List, Optional, Set, Tuple, Union
 
-from pycardano.address import Address
+from pycardano.address import Address, AddressType
 from pycardano.backend.base import ChainContext
+from pycardano.certificate import (
+    Certificate,
+    StakeCredential,
+    StakeDelegation,
+    StakeDeregistration,
+    StakeRegistration,
+)
 from pycardano.coinselection import (
     LargestFirstSelector,
     RandomImproveMultiAsset,
@@ -40,8 +47,9 @@ from pycardano.transaction import (
     TransactionOutput,
     UTxO,
     Value,
+    Withdrawals,
 )
-from pycardano.utils import fee, min_lovelace, script_data_hash
+from pycardano.utils import fee, max_tx_fee, min_lovelace, script_data_hash
 from pycardano.witness import TransactionWitnessSet, VerificationKeyWitness
 
 __all__ = ["TransactionBuilder"]
@@ -87,6 +95,10 @@ class TransactionBuilder:
 
     collaterals: List[UTxO] = field(default_factory=lambda: [])
 
+    certificates: List[Certificate] = field(default=None)
+
+    withdrawals: Withdrawals = field(default=None)
+
     _inputs: List[UTxO] = field(init=False, default_factory=lambda: [])
 
     _excluded_inputs: List[UTxO] = field(init=False, default_factory=lambda: [])
@@ -103,6 +115,10 @@ class TransactionBuilder:
         init=False, default_factory=lambda: {}
     )
 
+    _minting_script_to_redeemers: List[Tuple[bytes, Redeemer]] = field(
+        init=False, default_factory=lambda: []
+    )
+
     _inputs_to_scripts: Dict[UTxO, bytes] = field(
         init=False, default_factory=lambda: {}
     )
@@ -148,7 +164,11 @@ class TransactionBuilder:
                     redeemer.ex_units = ExecutionUnits(0, 0)
 
     def add_script_input(
-        self, utxo: UTxO, script: bytes, datum: Datum, redeemer: Redeemer
+        self,
+        utxo: UTxO,
+        script: bytes,
+        datum: Optional[Datum] = None,
+        redeemer: Optional[Redeemer] = None,
     ) -> TransactionBuilder:
         """Add a script UTxO to transaction's inputs.
 
@@ -166,18 +186,44 @@ class TransactionBuilder:
                 f"Expect the output address of utxo to be script type, "
                 f"but got {utxo.output.address.address_type} instead."
             )
-        if utxo.output.datum_hash != datum.hash():
+        if utxo.output.datum_hash and utxo.output.datum_hash != datum_hash(datum):
             raise InvalidArgumentException(
                 f"Datum hash in transaction output is {utxo.output.datum_hash}, "
-                f"but actual datum hash from input datum is {datum.hash()}."
+                f"but actual datum hash from input datum is {datum_hash(datum)}."
             )
-        self.datums[datum.hash()] = datum
-        self._consolidate_redeemer(redeemer)
-        self._inputs_to_redeemers[utxo] = redeemer
+        if datum:
+            self.datums[datum_hash(datum)] = datum
+        if redeemer:
+            self._consolidate_redeemer(redeemer)
+            self._inputs_to_redeemers[utxo] = redeemer
         self._inputs_to_scripts[utxo] = script
         self.inputs.append(utxo)
         return self
 
+    def add_minting_script(
+        self,
+        script: bytes,
+        redeemer: Optional[Redeemer] = None,
+    ) -> TransactionBuilder:
+        """Add a minting script along with its datum and redeemer to this transaction.
+
+        Args:
+            script (Optional[bytes]): A plutus script.
+            redeemer (Optional[Redeemer]): A plutus redeemer to unlock the UTxO.
+
+        Returns:
+            TransactionBuilder: Current transaction builder.
+        """
+        if redeemer:
+            if redeemer.tag != RedeemerTag.MINT:
+                raise InvalidArgumentException(
+                    f"Expect the redeemer tag's type to be {RedeemerTag.MINT}, "
+                    f"but got {redeemer.tag} instead."
+                )
+            self._consolidate_redeemer(redeemer)
+        self._minting_script_to_redeemers.append((script, redeemer))
+        return self
+
     def add_input_address(self, address: Union[Address, str]) -> TransactionBuilder:
         """Add an address to transaction's input address.
         Unlike :meth:`add_input`, which deterministically adds a UTxO to the transaction's inputs, `add_input_address`
@@ -213,7 +259,7 @@ class TransactionBuilder:
             tx_out.datum_hash = datum_hash(datum)
         self.outputs.append(tx_out)
         if add_datum_to_witness:
-            self.datums[datum.hash()] = datum
+            self.datums[datum_hash(datum)] = datum
         return self
 
     @property
@@ -246,7 +292,10 @@ class TransactionBuilder:
 
     @property
     def scripts(self) -> List[bytes]:
-        return list(set(self._inputs_to_scripts.values()))
+        return list(
+            set(self._inputs_to_scripts.values())
+            | {s for s, _ in self._minting_script_to_redeemers}
+        )
 
     @property
     def datums(self) -> Dict[DatumHash, Datum]:
@@ -254,7 +303,9 @@ class TransactionBuilder:
 
     @property
     def redeemers(self) -> List[Redeemer]:
-        return list(self._inputs_to_redeemers.values())
+        return list(self._inputs_to_redeemers.values()) + [
+            r for _, r in self._minting_script_to_redeemers
+        ]
 
     @property
     def script_data_hash(self) -> Optional[ScriptDataHash]:
@@ -275,6 +326,11 @@ class TransactionBuilder:
             provided += i.output.amount
         if self.mint:
             provided.multi_asset += self.mint
+        if self.withdrawals:
+            for v in self.withdrawals.values():
+                provided.coin += v
+
+        provided.coin -= self._get_total_key_deposit()
 
         if not requested < provided:
             raise InvalidTransactionException(
@@ -295,7 +351,9 @@ class TransactionBuilder:
         # when there is only ADA left, simply use remaining coin value as change
         if not change.multi_asset:
             if change.coin < min_lovelace(change, self.context):
-                raise InsufficientUTxOBalanceException("Not enough ADA left for change")
+                raise InsufficientUTxOBalanceException(
+                    f"Not enough ADA left for change: {change.coin} but needs {min_lovelace(change, self.context)}"
+                )
             lovelace_change = change.coin
             change_output_arr.append(TransactionOutput(address, lovelace_change))
 
@@ -330,17 +388,42 @@ class TransactionBuilder:
         return change_output_arr
 
     def _add_change_and_fee(
-        self, change_address: Optional[Address]
+        self,
+        change_address: Optional[Address],
+        merge_change: Optional[bool] = False,
     ) -> TransactionBuilder:
-        original_outputs = self.outputs[:]
+        original_outputs = deepcopy(self.outputs)
+        change_output_index = None
+
+        def _merge_changes(changes):
+            if change_output_index is not None and len(changes) == 1:
+                # Add the leftover change to the TransactionOutput containing the change address
+                self._outputs[change_output_index].amount = (
+                    changes[0].amount + self._outputs[change_output_index].amount
+                )
+                # if we enforce that TransactionOutputs must use Values for `amount`, we can use += here
+
+            else:
+                self._outputs += changes
 
         if change_address:
+
+            if merge_change:
+
+                for idx, output in enumerate(original_outputs):
+
+                    # Find any transaction outputs which already contain the change address
+                    if change_address == output.address:
+                        if change_output_index is None or output.lovelace == 0:
+                            change_output_index = idx
+
             # Set fee to max
             self.fee = self._estimate_fee()
             changes = self._calc_change(
                 self.fee, self.inputs, self.outputs, change_address, precise_fee=True
             )
-            self._outputs += changes
+
+            _merge_changes(changes)
 
         # With changes included, we can estimate the fee more precisely
         self.fee = self._estimate_fee()
@@ -350,7 +433,8 @@ class TransactionBuilder:
             changes = self._calc_change(
                 self.fee, self.inputs, self.outputs, change_address, precise_fee=True
             )
-            self._outputs += changes
+
+            _merge_changes(changes)
 
         return self
 
@@ -461,6 +545,42 @@ class TransactionBuilder:
                 results.add(i.output.address.payment_part)
         return results
 
+    def _certificate_vkey_hashes(self) -> Set[VerificationKeyHash]:
+
+        results = set()
+
+        def _check_and_add_vkey(stake_credential: StakeCredential):
+            if isinstance(stake_credential.credential, VerificationKeyHash):
+                results.add(stake_credential.credential)
+
+        if self.certificates:
+            for cert in self.certificates:
+                if isinstance(
+                    cert, (StakeRegistration, StakeDeregistration, StakeDelegation)
+                ):
+                    _check_and_add_vkey(cert.stake_credential)
+        return results
+
+    def _get_total_key_deposit(self):
+        results = set()
+        if self.certificates:
+            for cert in self.certificates:
+                if isinstance(cert, StakeRegistration):
+                    results.add(cert.stake_credential.credential)
+        return self.context.protocol_param.key_deposit * len(results)
+
+    def _withdrawal_vkey_hashes(self) -> Set[VerificationKeyHash]:
+
+        results = set()
+
+        if self.withdrawals:
+            for k in self.withdrawals:
+                address = Address.from_primitive(k)
+                if address.address_type == AddressType.NONE_KEY:
+                    results.add(address.staking_part)
+
+        return results
+
     def _native_scripts_vkey_hashes(self) -> Set[VerificationKeyHash]:
         results = set()
 
@@ -504,6 +624,10 @@ class TransactionBuilder:
                 redeemer.index = sorted_mint_policies.index(
                     plutus_script_hash(self._inputs_to_scripts[utxo])
                 )
+
+        for script, redeemer in self._minting_script_to_redeemers:
+            redeemer.index = sorted_mint_policies.index(plutus_script_hash(script))
+
         self.redeemers.sort(key=lambda r: r.index)
 
     def _build_tx_body(self) -> TransactionBody:
@@ -522,12 +646,16 @@ class TransactionBuilder:
             collateral=[c.input for c in self.collaterals]
             if self.collaterals
             else None,
+            certificates=self.certificates,
+            withdraws=self.withdrawals,
         )
         return tx_body
 
     def _build_fake_vkey_witnesses(self) -> List[VerificationKeyWitness]:
         vkey_hashes = self._input_vkey_hashes()
         vkey_hashes.update(self._native_scripts_vkey_hashes())
+        vkey_hashes.update(self._certificate_vkey_hashes())
+        vkey_hashes.update(self._withdrawal_vkey_hashes())
         return [
             VerificationKeyWitness(FAKE_VKEY, FAKE_TX_SIGNATURE) for _ in vkey_hashes
         ]
@@ -539,6 +667,12 @@ class TransactionBuilder:
 
     def _build_full_fake_tx(self) -> Transaction:
         tx_body = self._build_tx_body()
+
+        if tx_body.fee == 0:
+            # When fee is not specified, we will use max possible fee to fill in the fee field.
+            # This will make sure the size of fee field itself is taken into account during fee estimation.
+            tx_body.fee = max_tx_fee(self.context)
+
         witness = self._build_fake_witness_set()
         tx = Transaction(tx_body, witness, True, self.auxiliary_data)
         size = len(tx.to_cbor("bytes"))
@@ -586,12 +720,18 @@ class TransactionBuilder:
 
         return estimated_fee
 
-    def build(self, change_address: Optional[Address] = None) -> TransactionBody:
+    def build(
+        self,
+        change_address: Optional[Address] = None,
+        merge_change: Optional[bool] = False,
+    ) -> TransactionBody:
         """Build a transaction body from all constraints set through the builder.
 
         Args:
             change_address (Optional[Address]): Address to which changes will be returned. If not provided, the
                 transaction body will likely be unbalanced (sum of inputs is greater than the sum of outputs).
+            merge_change (Optional[bool]): If the change address match one of the transaction output, the change amount
+                will be directly added to that transaction output, instead of being added as a separate output.
 
         Returns:
             TransactionBody: A transaction body.
@@ -603,9 +743,23 @@ class TransactionBuilder:
         for i in self.inputs:
             selected_utxos.append(i)
             selected_amount += i.output.amount
+
         if self.mint:
             selected_amount.multi_asset += self.mint
 
+        if self.withdrawals:
+            for v in self.withdrawals.values():
+                selected_amount.coin += v
+
+        can_merge_change = False
+        if merge_change:
+            for o in self.outputs:
+                if o.address == change_address:
+                    can_merge_change = True
+                    break
+
+        selected_amount.coin -= self._get_total_key_deposit()
+
         requested_amount = Value()
         for o in self.outputs:
             requested_amount += o.amount
@@ -624,16 +778,20 @@ class TransactionBuilder:
 
         unfulfilled_amount = requested_amount - trimmed_selected_amount
 
-        if change_address is not None:
+        if change_address is not None and not can_merge_change:
             # If change address is provided and remainder is smaller than minimum ADA required in change,
             # we need to select additional UTxOs available from the address
-            unfulfilled_amount.coin = max(
-                0,
-                unfulfilled_amount.coin
-                + min_lovelace(selected_amount - trimmed_selected_amount, self.context),
-            )
+            if unfulfilled_amount.coin < 0:
+                unfulfilled_amount.coin = max(
+                    0,
+                    unfulfilled_amount.coin
+                    + min_lovelace(
+                        selected_amount - trimmed_selected_amount, self.context
+                    ),
+                )
         else:
             unfulfilled_amount.coin = max(0, unfulfilled_amount.coin)
+
         # Clean up all non-positive assets
         unfulfilled_amount.multi_asset = unfulfilled_amount.multi_asset.filter(
             lambda p, n, v: v > 0
@@ -660,6 +818,7 @@ class TransactionBuilder:
                         [TransactionOutput(None, unfulfilled_amount)],
                         self.context,
                         include_max_fee=False,
+                        respect_min_utxo=not can_merge_change,
                     )
                     for s in selected:
                         selected_amount += s.output.amount
@@ -703,7 +862,7 @@ class TransactionBuilder:
 
         self._set_redeemer_index()
 
-        self._add_change_and_fee(change_address)
+        self._add_change_and_fee(change_address, merge_change=merge_change)
 
         tx_body = self._build_tx_body()
 
@@ -750,6 +909,7 @@ class TransactionBuilder:
         self,
         signing_keys: List[Union[SigningKey, ExtendedSigningKey]],
         change_address: Optional[Address] = None,
+        merge_change: Optional[bool] = False,
     ) -> Transaction:
         """Build a transaction body from all constraints set through the builder and sign the transaction with
         provided signing keys.
@@ -759,12 +919,14 @@ class TransactionBuilder:
                 sign the transaction.
             change_address (Optional[Address]): Address to which changes will be returned. If not provided, the
                 transaction body will likely be unbalanced (sum of inputs is greater than the sum of outputs).
+            merge_change (Optional[bool]): If the change address match one of the transaction output, the change amount
+                will be directly added to that transaction output, instead of being added as a separate output.
 
         Returns:
             Transaction: A signed transaction.
         """
 
-        tx_body = self.build(change_address=change_address)
+        tx_body = self.build(change_address=change_address, merge_change=merge_change)
         witness_set = self.build_witness_set()
         witness_set.vkey_witnesses = []
 
